#pragma once
#include <stdlib.h>
#include <math.h>
#include "PPInterpolatable.h"
#include "TensorInterpolatable.h"



template<typename IP> // must be either PPInterpolatable or TensorInterpolatable, or any class with a similar class structure 
class Diffusion {
public:

	static enum DIFFUSION_METHOD {
		EXPLICIT, 
		GAUSS_SEIDEL, 
	};

	Diffusion(){ 
		use_pyramid = false;
		setBlurMethod(GAUSS_SEIDEL);
		enableConvergenceCheck(false);
		converged = true;
		conv_criteria = 0.01f;
		gs_conv_itr = 20;
		error.zero();
	}

	~Diffusion(){}


	void setBlurMethod(typename Diffusion::DIFFUSION_METHOD method) { 
		switch(method) {
			case EXPLICIT:
				current_method = method;
				//_blur = &PSFBlur::blur_ex;
				break;

			case GAUSS_SEIDEL:
				current_method = method;
				//_blur = &PSFBlur::blur_gs;
				break;
		}
	}

	typename Diffusion::DIFFUSION_METHOD getBlurMethod() { return current_method; }

	void blurWithPyramid(bool val) { use_pyramid = val; }
	bool isPyramidEnabled() { return use_pyramid; }

	void setConvergenceCriteria(float val) { conv_criteria = val; }
	float getConvergenceCriteria() { return conv_criteria; }

	void setGSConvergenceIterations(int val) { gs_conv_itr = val; }
	int getGSConvergenceIterations() { return gs_conv_itr; }


	bool isConverged() { return converged; }
	IP getMaxError() { return error; }


	// dt = 1.f is preferrable with Gauss-Seidel with Pyramid
	void blur(IP *arr, int width, int height, float dt, int num_itr = 1, IP *buffer1 = NULL) {
		(use_pyramid == false) ? 
			_blur(arr, width, height, dt, num_itr, buffer1) 
			: blurPyramid(arr, width, height, dt, num_itr, buffer1);
	}

	void enableConvergenceCheck(bool val) {
		conv_check = val;
	}

	bool isConvergenceCheckEnabled() { return conv_check; }

private:
	bool use_pyramid;
	DIFFUSION_METHOD current_method;

	bool converged;
	bool conv_check;
	IP error;
	float conv_criteria;		// threshold to determine heat diffusion is in equibilium state

	// specific to Gauss-Seidel method...
	int gs_conv_itr;

	// function pointer
	//void (Diffusion::*_blur)(IP *arr, int width, int height, float dt, int num_itr, PPInterpolatable *b1);

	/**/
	// how can I make a function pointer of a template class?
	void _blur(IP *arr, int width, int height, float dt, int num_itr, IP *b1) {
		switch(current_method) {
			case EXPLICIT:
			default:
				blur_ex(arr, width, height, dt, num_itr, b1);
				break;

			case GAUSS_SEIDEL:
				blur_gs(arr, width, height, dt, num_itr, b1);
				break;
		}
	}
	/**/



	// return the exponent of val if POT, -1 otherwise.
	inline int checkPOT(unsigned int val) {
		int cnt = 0;
		unsigned int tmp = 1;


		while(tmp < val) {
			cnt++;
			tmp *= 2;
		}

		return (tmp == val) ? cnt : -1;
	}




	// return the highest pyramid level to use
	// it assumes that inputs are all valid... 
	int push(IP **pyr, int width, int height, int levels) {
		int prev_w, prev_h, prev_idx;
		int curr_w, curr_h, curr_idx;
		int num_elems;
		int uplimit_level = 0;
		int cnt;
		bool flag;							// to push values to deeper levels...
		float ooc;
		IP acc;



		flag = false;
		num_elems = width * height;

		for(int i = 0; i < num_elems; i++) {
			if(pyr[0][i].mask == true) {
				flag = true;
				break;
			}
		}


		// all values are defined so no need to push.
		if(flag == false) return uplimit_level;



		// lower level
		prev_w = width;
		prev_h = height;

		// higher level
		curr_w = prev_w >> 1;
		curr_h = prev_h >> 1;


		uplimit_level++;

		for(int i = 1; i < levels && flag == true; i++) {
			flag = false;

			for(int j = 0; j < curr_h; j++) {
				for(int k = 0; k < curr_w; k++) {
					curr_idx = k + curr_w * j;
					cnt = 0;
					acc.zero();

					for(int m = 0; m < 2; m++) {
						for(int n = 0; n < 2; n++) {
							prev_idx = ((k << 1) + n) + prev_w * ((j << 1) + m);

							if(pyr[i - 1][prev_idx].mask == false) {
								cnt++;
								acc += pyr[i - 1][prev_idx];
							}
						}
					}

					if(cnt == 0) {
						flag = true;		// need to propagate deeper.

						pyr[i][curr_idx].mask = true;
					}
					else {
						ooc = 1.f / cnt;

						pyr[i][curr_idx].mask = false;
						pyr[i][curr_idx] = acc * ooc;
					}
				}
			}

			if(flag == true) {
				curr_w = curr_w >> 1;
				curr_h = curr_h >> 1;
				prev_w = prev_w >> 1;
				prev_h = prev_h >> 1;
				uplimit_level++;
			}
		}

		if(uplimit_level >= levels) uplimit_level = levels - 1;

		return uplimit_level;
	}




	// it assumes that inputs are all valid... 
	void pull(IP **pyr, int width, int height, int levels, float dt, int num_itr, IP *b1) {
		int itr_cnt = 0;
		int max_itr_limit = 500;
		int prev_w, prev_h, prev_idx;
		int curr_w, curr_h, curr_idx;


		prev_w = width;
		prev_h = height;

		for(int i = 0; i < levels; i++) {
			prev_w = prev_w >> 1;
			prev_h = prev_h >> 1;
		}

		curr_w = prev_w << 1;
		curr_h = prev_h << 1;


		for(int i = levels - 1; i >= 0; i--) {
			for(int j = 0; j < prev_h; j++) {
				for(int k = 0; k < prev_w; k++) {
					prev_idx = k + prev_w * j;

					for(int m = 0; m < 2; m++) {
						for(int n = 0; n < 2; n++) {
							curr_idx = ((k << 1) + n) + curr_w * ((j << 1) + m);

							if(pyr[i][curr_idx].mask == true) {
								pyr[i][curr_idx] = pyr[i + 1][prev_idx];
							}
						}
					}
				}
			}

			if(conv_check == true) {
				itr_cnt = 0;

				do {
					// blur for some number of iterations...
					//(this->*_blur)(pyr[i], curr_w, curr_h, dt, num_itr, b1);
					_blur(pyr[i], curr_w, curr_h, dt, num_itr, b1);

					// avoid infinite loop
					if(itr_cnt++ > max_itr_limit) {
						printf("iterations exceed limit.\n");
						break;
					}
				}
				while(isConverged() == false);
			}
			else {
				//(this->*_blur)(pyr[i], curr_w, curr_h, dt, num_itr, b1);
				_blur(pyr[i], curr_w, curr_h, dt, num_itr, b1);
			}


			prev_w = prev_w << 1;
			prev_h = prev_h << 1;
			curr_w = curr_w << 1;
			curr_h = curr_h << 1;
		}
	}




	// assume that input array dimensions are POT
	void blurPyramid(IP *arr, int width, int height, float dt, int num_itr, IP *b1) {
		int num_elems = width * height;
		int num_levels;
		int lvl_w, lvl_h;
		int uplimit_level;
		IP **pyr;
		IP *_b1;


		if(width <= 0 || height <= 0 || num_elems == 1) return;

		// return if npot.
		if((lvl_w = checkPOT(width)) < 0) return;
		if((lvl_h = checkPOT(height)) < 0) return;

		lvl_w++;
		lvl_h++;


		_b1 = (b1 == NULL) ? new IP[width * height] : b1;


		// build a pyramid
		num_elems = width * height;
		num_levels = (lvl_w < lvl_h) ? lvl_w : lvl_h;		// take whichever smaller
		pyr = new IP*[num_levels];
		pyr[0] = arr;


		for(int i = 1; i < num_levels; i++) {
			pyr[i] = new IP[(int)(num_elems >> (2 * i))];
		}

		uplimit_level = push(pyr, width, height, num_levels);

		pull(pyr, width, height, uplimit_level, dt, num_itr, b1);

		for(int i = 1; i < num_levels; i++) {
			delete[] pyr[i];
		}

		delete[] pyr;

		if(b1 == NULL) delete[] _b1;
	}



	
	void doConvCheck(IP *arr, int width, int height) {
		int idx1;		// target
		int ng[4];		// neighbors
		int num_elems = width * height;
		IP acc;


		converged = true;
		error.zero();

		for(int i = 0; i < height; i++) {
			for(int j = 0; j < width; j++) {
				idx1 = j + width * i;

				// we skip elements not necessary to be computed
				if(arr[idx1].mask == false) continue;

				// compute neighbors
				ng[0] = idx1 + ((j < width - 1)		? 1		: 0);		// right
				ng[1] = idx1 + ((i < height - 1)	? width : 0);		// top
				ng[2] = idx1 - ((j > 0)				? 1		: 0);		// left
				ng[3] = idx1 - ((i > 0)				? width : 0);		// right

				// compute Laplacian of the current solution.
				acc = arr[ng[0]] + arr[ng[1]] + arr[ng[2]] + arr[ng[3]] - 4 * arr[idx1];

				// take max. of each member
				if(acc > error) error._max(acc);
				if(error > conv_criteria) converged = false;
			}
		}
	}




	void blur_gs(IP *arr, int width, int height, float dt, int num_itr, IP *b1) {
		int idx1;		// target
		int ng[4];		// neighbors
		int num_elems = width * height;
		float denom;
		IP acc;
		IP *prev = NULL;



		if(width <= 0 || height <= 0 || num_elems == 1) return;

		prev = (b1 == NULL) ? new IP[width * height] : b1;
		denom = 1.f / (1 + 4 * dt);


		for(int g = 0; g < num_itr; g++) {
			for(int i = 0; i < num_elems; i++) {
				prev[i] = arr[i];
			}

			// we omit error checking at each iteration, assuming it converges within a specified iterations.
			for(int h = 0; h < gs_conv_itr; h++) {
				for(int i = 0; i < height; i++) {
					for(int j = 0; j < width; j++) {
						idx1 = j + width * i;

						// we skip elements not necessary to be computed
						if(arr[idx1].mask == false) continue;


						// compute neighbors
						ng[0] = idx1 + ((j < width - 1)		? 1		: 0);		// right
						ng[1] = idx1 + ((i < height - 1)	? width : 0);		// top
						ng[2] = idx1 - ((j > 0)				? 1		: 0);		// left
						ng[3] = idx1 - ((i > 0)				? width : 0);		// right

						acc  = arr[ng[0]];
						acc += arr[ng[1]];
						acc += arr[ng[2]];
						acc += arr[ng[3]];
						arr[idx1] = (prev[idx1] + acc * dt) * denom;
					}
				}
			}

			if(conv_check == true) {
				doConvCheck(arr, width, height);
			}
		}


		if(b1 == NULL) delete[] prev;
	}




	// Explicit method
	void blur_ex(IP *arr, int width, int height, float dt, int num_itr, IP *b1) {
		int idx1;		// target
		int ng[4];		// neighbors
		int num_elems = width * height;
		IP acc;
		IP *prev = NULL;


		if(width <= 0 || height <= 0 || num_elems == 1) return;


		prev = (b1 == NULL) ? new IP[width * height] : b1;

		for(int h = 0; h < num_itr; h++) {
			for(int i = 0; i < num_elems; i++) {
				prev[i] = arr[i];
			}

			for(int i = 0; i < height; i++) {
				for(int j = 0; j < width; j++) {
					idx1 = j + width * i;

					// we skip elements not necessary to be computed
					if(arr[idx1].mask == false) continue;

					// compute neighbors
					ng[0] = idx1 + ((j < width - 1)		? 1		: 0);		// right
					ng[1] = idx1 + ((i < height - 1)	? width : 0);		// top
					ng[2] = idx1 - ((j > 0)				? 1		: 0);		// left
					ng[3] = idx1 - ((i > 0)				? width : 0);		// right

					acc = prev[ng[0]] + prev[ng[1]] + prev[ng[2]] + prev[ng[3]] - 4 * prev[idx1];
					arr[idx1] += acc * dt;
				}
			}

			if(conv_check == true) {
				doConvCheck(arr, width, height);
			}
		}


		if(b1 == NULL) delete[] prev;
	}
};

