#pragma once
#include "StdAfx.h"
#include "EdgeFieldConstructor.h"
#include "cv.h"
#include "cvaux.h"
#include "cxcore.h"
#include "highgui.h"


#define USE_OMP 1

#ifndef _OPENMP
#define USE_OMP 0
#endif

#if USE_OMP == 1
#include <omp.h>
#endif



const double EdgeFieldConstructor::gaussian1D[5] = {.05, .25, .40, .25, .05};


bool EdgeFieldConstructor::computeEdgeField(unsigned char *(&img), FieldSampleImage &fsi, int num_blur) {
	int num_elems;
	int num_pixels, num_pixels2;
	int new_width, new_height;
	double max_vec_length, temp_vec_length, oomvl;
	double len, angle, ctest, stest;
	IplImage *ipl_input, *ipl_gray;
	CvMat *sobelx, *sobely;

	char *img_ptr;
	double *vec_entry;
	icVector2 *vec;
	icMatrix2x2 *ten;

	clock_t t1[2], t2[2];



	//int apature = -1;		// 3x3 Scharr filter...
	int apature = 3;		// 3x3 Sobel filter


	if(fsi.w <= 0 || fsi.h <= 0) return false;


	t1[0] = clock();


	num_pixels = fsi.w * fsi.h;
	num_elems = 3 * num_pixels;

	new_width = fsi.w + 2;
	new_height = fsi.h + 2;
	num_pixels2 = new_width * new_height;

	ipl_input = cvCreateImage(cvSize(fsi.w, fsi.h), IPL_DEPTH_8U, 3);
	ipl_gray = cvCreateImage(cvSize(fsi.w, fsi.h), IPL_DEPTH_8U, 1);


	// copy
	img_ptr = ipl_input->imageDataOrigin;

	for(int i = 0; i < num_elems; i++) {
		img_ptr[i] = static_cast<char>(img[i]);
	}


	// convert to the 'BGR' color mode
	cvConvertImage(ipl_input, ipl_input, CV_CVTIMG_SWAP_RB);

	for(int i = 0; i < num_blur; i++) {
		cvSmooth(ipl_input, ipl_input, CV_GAUSSIAN, 3, 0);
	}

	cvConvertImage(ipl_input, ipl_gray, CV_BGR2GRAY);
	cvReleaseImage(&ipl_input);


	sobelx = cvCreateMat(fsi.h, fsi.w, CV_16S);
	sobely = cvCreateMat(fsi.h, fsi.w, CV_16S);
	cvSobel(ipl_gray, sobelx, 1, 0, apature);
	cvSobel(ipl_gray, sobely, 0, 1, apature);
	cvReleaseImage(&ipl_gray);

	max_vec_length = 0;

	t2[0] = clock();
	t1[1] = clock();

#if USE_OMP == 1
	double max_vec_len_per_thread = 0;

	#pragma omp parallel firstprivate(max_vec_len_per_thread) private(vec_entry, temp_vec_length, oomvl, vec, ten, len, angle, ctest, stest)
	{
		#pragma omp for
		for(int i = 0; i < num_pixels; i++) {
			vec_entry = &(fsi.fld[i].vec_fld.entry[0]);
			vec_entry[0] = sobely->data.s[i];
			vec_entry[1] = sobelx->data.s[i];
			temp_vec_length = vec_entry[0] * vec_entry[0] + vec_entry[1] * vec_entry[1];

			if(max_vec_len_per_thread < temp_vec_length)
				max_vec_len_per_thread = temp_vec_length;
		}

		#pragma omp critical (get_min)
		{
			if(max_vec_length < max_vec_len_per_thread) {
				max_vec_length = max_vec_len_per_thread;
			}
		}

		#pragma omp barrier

		#pragma omp single
		{
			max_vec_length = sqrt(max_vec_length);
		}

		// scale the gradient so to be within [0, 1]

		if(max_vec_length > 0) {
			oomvl = 1. / max_vec_length;

			#pragma omp for
			for(int i = 0; i < num_pixels; i++) {
				vec = &(fsi.fld[i].vec_fld);
				*vec *= oomvl;
				len = length(*vec);
				angle = 2. * atan2(vec->entry[1], vec->entry[0]);
				ctest = cos(angle) * len;
				stest = sin(angle) * len;

				ten = &(fsi.fld[i].ten_fld);
				ten->entry[0][0] = ctest;
				ten->entry[0][1] = stest;
				ten->entry[1][0] = stest;
				ten->entry[1][1] = -ctest;
			}
		}
		else {
			#pragma omp for
			for(int i = 0; i < num_pixels; i++) {
				fsi.fld[i].vec_fld.set(0.0);
				fsi.fld[i].ten_fld.set(0.0);
			}
		}
	}

#else

	for(int i = 0; i < num_pixels; i++) {
		vec_entry = &(fsi.fld[i].vec_fld.entry[0]);
		vec_entry[0] = sobely->data.s[i];
		vec_entry[1] = sobelx->data.s[i];
		temp_vec_length = vec_entry[0] * vec_entry[0] + vec_entry[1] * vec_entry[1];

		
		if(max_vec_length < temp_vec_length)
			max_vec_length = temp_vec_length;
	}

	max_vec_length = sqrt(max_vec_length);

	// scale the gradient so to be within [0, 1]

	if(max_vec_length > 0) {
		oomvl = 1. / max_vec_length;

		for(int i = 0; i < num_pixels; i++) {
			vec = &(fsi.fld[i].vec_fld);
			*vec *= oomvl;
			len = length(*vec);
			angle = 2. * atan2(vec->entry[1], vec->entry[0]);
			ctest = cos(angle) * len;
			stest = sin(angle) * len;

			ten = &(fsi.fld[i].ten_fld);
			ten->entry[0][0] = ctest;
			ten->entry[0][1] = stest;
			ten->entry[1][0] = stest;
			ten->entry[1][1] = -ctest;
		}
	}
	else {
		for(int i = 0; i < num_pixels; i++) {
			fsi.fld[i].vec_fld.set(0.0);
			fsi.fld[i].ten_fld.set(0.0);
		}
	}


#endif


	cvReleaseMat(&sobelx);
	cvReleaseMat(&sobely);

	t2[1] = clock();

	/**
	FILE *fp = fopen("time.txt", "a+");
	fprintf(fp, "%.3lf, %.3lf, %.3lf\n", max_vec_length, (double)(t2[0]-t1[0])/CLOCKS_PER_SEC, (double)(t2[1]-t1[1])/CLOCKS_PER_SEC);
	fclose(fp);
	/**/

	return true;
}

/*
	angle_constraint_radian: assumed to be between 0 and PI/2
*/
bool EdgeFieldConstructor::refineEdgeField(FieldSampleImage &target, const FieldSampleImage &src, float angle_constraint_radian, bool recompute_vector) {
	int w, h, num_pixels;
	int sign;
	float ip_vecs, op_vecs;
	float ip_constraint;
	float cosine, sine;
	double vec00, vec01, vec10, vec11;
	double len0, len1, angle, ten_c, ten_s;
	double two_inv_len;
	FieldSample *fs0, *fs1;
	icVector2 n0, n1;


	if(target.w != src.w || target.h != src.h) return false;

	w = target.w;
	h = target.h;
	num_pixels = w * h;
	cosine = cos(angle_constraint_radian);
	sine = sin(angle_constraint_radian);
	ip_constraint = fabs(cosine);

#if USE_OMP == 1
//#if 0
	#pragma omp for private(fs0, fs1, vec00, vec01, vec10, vec11, n0, n1, ip_vecs, op_vecs, sign, ten_c, ten_s, two_inv_len, len0, len1, angle)
#endif
	for(int i = 0; i < num_pixels; i++) {
		fs0 = &(target.fld[i]);
		fs1 = &(src.fld[i]);

		// I assume that each tensor has symmetric form  of (r * [a, b; b, -a])...
		len0 = fs0->ten_fld.entry[0][0] * fs0->ten_fld.entry[0][0] + fs0->ten_fld.entry[0][1] * fs0->ten_fld.entry[0][1];

		if(len0 > 0) {
			len0 = sqrt(len0);
			angle = atan2(fs0->ten_fld.entry[0][1], fs0->ten_fld.entry[0][0]) * .5;
			n0.set(cos(angle), sin(angle));
			vec00 = n0.entry[0] * len0;
			vec01 = n0.entry[1] * len0;
		}
		else {
			n0.set(0, 0);
			vec00 = 0;
			vec01 = 0;
		}


		len1 = fs1->ten_fld.entry[0][0] * fs1->ten_fld.entry[0][0] + fs1->ten_fld.entry[0][1] * fs1->ten_fld.entry[0][1];

		if(len1 > 0) {
			len1 = sqrt(len1);
			angle = atan2(fs1->ten_fld.entry[0][1], fs1->ten_fld.entry[0][0]) * .5;
			n1.set(cos(angle), sin(angle));
			vec10 = n1.entry[0] * len1;
			vec11 = n1.entry[1] * len1;
		}
		else {
			n1.set(0, 0);
			vec10 = 0;
			vec11 = 0;
		}

		// measure angle difference between major eigenvectors (in terms of the dot product of the two)
		if((ip_vecs = (float)dot(n0, n1)) < 0) ip_vecs = (float)dot(n0, -n1);

		// if out of constraint, set it, and then rebulid a tensor.
		if(len1 > 0 && ip_vecs < ip_constraint) {
			op_vecs = (float)(vec10 * vec01 - vec11 * vec00);		// v1 X v0

			if(op_vecs < 0) sign = -1;
			else if(op_vecs > 0) sign = 1;
			else sign = 1;

			vec00 = cosine * vec10 - sign * sine * vec11;
			vec01 = sign * sine * vec10 + cosine * vec11;

			angle = atan2(vec01, vec00) * 2.;
			ten_c = len0 * cos(angle);
			ten_s = len0 * sin(angle);
			fs0->ten_fld.set(ten_c, ten_s, ten_s, -ten_c);
		}
	}


	if(recompute_vector == true) {

#if USE_OMP == 1
//#if 0
		#pragma omp parallel for private(fs0, len0, angle)
#endif
		for(int i = 0; i < num_pixels; i++) {
			fs0 = &(target.fld[i]);
			len0 = fs0->ten_fld.entry[0][0] * fs0->ten_fld.entry[0][0] + fs0->ten_fld.entry[0][1] * fs0->ten_fld.entry[0][1];

			if(len0 > 0) {
				len0 = sqrt(len0);
				angle = atan2(fs0->ten_fld.entry[0][1], fs0->ten_fld.entry[0][0]) / 2;
			}
			else {
				angle = 0;
			}

			fs0->vec_fld.entry[0] = cos(angle) * len0;
			fs0->vec_fld.entry[1] = sin(angle) * len0;
		}
	}

	return true;
}


bool EdgeFieldConstructor::averageEdgeField_S(FieldSampleImage &fsi, bool recompute_vector) {
	int idx, idx2;
	int num_pixels;
	double len, angle;
	icMatrix2x2 *mat;
	FieldSample *fs;
	FieldSampleImage *tmp;



	if(fsi.w <= 0 || fsi.h <= 0) return false;


	tmp = new FieldSampleImage(fsi.w + 4, fsi.h + 4);
	mat = new icMatrix2x2[tmp->w * tmp->h];
	num_pixels = fsi.w * fsi.h;

	addPadding(&fsi, tmp);


#if USE_OMP == 1
	#pragma omp parallel private(fs)
	{
#endif

	// convolution in x-direction
#if USE_OMP == 1
	#pragma omp for private(idx)
#endif
	for(int i = 0; i < tmp->h; i++) {
		idx = 2 + tmp->w * i;

		for(int j = 2; j < tmp->w - 2; j++) {
			mat[idx] = (gaussian1D[0] * tmp->fld[idx - 2].ten_fld);
			mat[idx] += (gaussian1D[1] * tmp->fld[idx - 1].ten_fld);
			mat[idx] += (gaussian1D[2] * tmp->fld[idx].ten_fld);
			mat[idx] += (gaussian1D[3] * tmp->fld[idx + 1].ten_fld);
			mat[idx] += (gaussian1D[4] * tmp->fld[idx + 2].ten_fld);

			idx++;
		}
	}

	// convolution in y-direction
#if USE_OMP == 1
	#pragma omp for private(idx, idx2)
#endif
	for(int i = 2; i < tmp->w - 2; i++) {
		idx = i + tmp->w * 2;
		idx2 = i - 2;

		for(int j = 2; j < tmp->h - 2; j++) {
			fs = &(fsi.fld[idx2]);
			fs->ten_fld = (gaussian1D[0] * mat[idx - 2 * tmp->w]);
			fs->ten_fld += (gaussian1D[1] * mat[idx - tmp->w]);
			fs->ten_fld += (gaussian1D[2] * mat[idx]);
			fs->ten_fld += (gaussian1D[3] * mat[idx + tmp->w]);
			fs->ten_fld += (gaussian1D[4] * mat[idx + 2 * tmp->w]);

			idx += tmp->w;
			idx2 += fsi.w;
		}
	}

#if USE_OMP == 1
	}
#endif


	delete[] mat;
	tmp->finalize();


	// I assume that each tensor has symmetric form  of (r * [a, b; b, -a])...
	if(recompute_vector == true) {

#if USE_OMP == 1
		#pragma omp parallel for private(fs, len, angle)
#endif
		for(int i = 0; i < num_pixels; i++) {
			fs = &(fsi.fld[i]);
			len = fs->ten_fld.entry[0][0] * fs->ten_fld.entry[0][0] + fs->ten_fld.entry[0][1] * fs->ten_fld.entry[0][1];

			if(len > 0) {
				len = sqrt(len);
				angle = atan2(fs->ten_fld.entry[0][1], fs->ten_fld.entry[0][0]) / 2;
			}
			else {
				angle = 0;
			}

			fs->vec_fld.entry[0] = cos(angle) * len;
			fs->vec_fld.entry[1] = sin(angle) * len;
		}
	}

	return true;
}



bool EdgeFieldConstructor::averageEdgeField_T(const list<FieldSampleImage *> &fsi_list, FieldSampleImage &out, bool recompute_vector) {
	list<FieldSampleImage*>::const_iterator it;
	double len, angle;
	FieldSample *fs;
	int num_pixels;
	int fsi_count;


	if(fsi_list.size() != temporal_kernel_size) return false;		// restrict to 5 frames...
	if(out.w <= 0 || out.h <= 0) return false;

	for(it = fsi_list.begin(); it != fsi_list.end(); it++) {
		if((*it)->w != out.w || (*it)->h != out.h) return false;
	}


	num_pixels = out.w * out.h;
	fsi_count = 0;
	it = fsi_list.begin();


	// Average tensor
#if USE_OMP == 1
	#pragma omp parallel private(fs)
	{
		#pragma omp for
		for(int i = 0; i < num_pixels; i++) {
			fs = &(out.fld[i]);
			fs->ten_fld = (gaussian1D[fsi_count] * (*it)->fld[i].ten_fld);
		}

		#pragma omp single
		{
			it++;
			fsi_count++;
		}

		while(it != fsi_list.end()) {
			#pragma omp for
			for(int i = 0; i < num_pixels; i++) {
				fs = &(out.fld[i]);
				fs->ten_fld += (gaussian1D[fsi_count] * (*it)->fld[i].ten_fld);
			}

			#pragma omp single
			{
				it++;
				fsi_count++;
			}
		}
	}

#else
	for(int i = 0; i < num_pixels; i++) {
		fs = &(out.fld[i]);
		fs->ten_fld = (gaussian1D[fsi_count] * (*it)->fld[i].ten_fld);
	}

	it++;
	fsi_count++;

	while(it != fsi_list.end()) {
		for(int i = 0; i < num_pixels; i++) {
			fs = &(out.fld[i]);
			fs->ten_fld += (gaussian1D[fsi_count] * (*it)->fld[i].ten_fld);
		}

		it++;
		fsi_count++;
	}

#endif



	// I assume that each tensor has symmetric form  of (r * [a, b; b, -a])...
	if(recompute_vector == true) {

#if USE_OMP == 1
		#pragma omp parallel for private(fs, len, angle)
#endif
		for(int i = 0; i < num_pixels; i++) {
			fs = &(out.fld[i]);
			len = fs->ten_fld.entry[0][0] * fs->ten_fld.entry[0][0] + fs->ten_fld.entry[0][1] * fs->ten_fld.entry[0][1];

			if(len > 0) {
				len = sqrt(len);
				angle = atan2(fs->ten_fld.entry[0][1], fs->ten_fld.entry[0][0]) / 2;
			}
			else {
				angle = 0;
			}

			fs->vec_fld.entry[0] = cos(angle) * len;
			fs->vec_fld.entry[1] = sin(angle) * len;
		}
	}

	return true;
}

/*
  Given (w * h) FieldSampleImage, this function creates (w + 4) * (h + 4) FieldSampleImage by padding new elements on sides.
  Values of each added element are computed by a simple extrapolation.
*/
void EdgeFieldConstructor::addPadding(const FieldSampleImage *in, FieldSampleImage *out)  {
	int idx1, idx2, idx3;
	int w, h;


	w = in->w;
	h = in->h;


#if USE_OMP == 1
	#pragma omp parallel private(idx1, idx2, idx3)
	{
#endif

	// inside
#if USE_OMP == 1
	#pragma omp for
#endif
	for(int i = 0; i < h; i++) {
		idx1 = w * i;
		idx2 = 2 + (w + 4) * (i + 2);

		for(int j = 0; j < w; j++) {
			out->fld[idx2].vec_fld = in->fld[idx1].vec_fld;
			out->fld[idx2++].ten_fld = in->fld[idx1++].ten_fld;
		}
	}


	// left & right
#if USE_OMP == 1
	#pragma omp for
#endif
	for(int i = 2; i < h + 2; i++) {
		idx1 = (w + 4) * i + 1;
		idx2 = (w + 4) * i + 2;
		idx3 = (w + 4) * i + 3;
		out->fld[idx1].ten_fld = 2 * out->fld[idx2].ten_fld - out->fld[idx3].ten_fld;

		idx1 = (w + 4) * i + 0;
		idx3 = (w + 4) * i + 4;
		out->fld[idx1].ten_fld = 2 * out->fld[idx2].ten_fld - out->fld[idx3].ten_fld;

		idx1 = (w + 4) * (i + 1) - 2;
		idx2 = (w + 4) * (i + 1) - 3;
		idx3 = (w + 4) * (i + 1) - 4;
		out->fld[idx1].ten_fld = 2 * out->fld[idx2].ten_fld - out->fld[idx3].ten_fld;

		idx1 = (w + 4) * (i + 1) - 1;
		idx3 = (w + 4) * (i + 1) - 5;
		out->fld[idx1].ten_fld = 2 * out->fld[idx2].ten_fld - out->fld[idx3].ten_fld;
	}


	// top & bottom
#if USE_OMP == 1
	#pragma omp for
#endif
	for(int i = 2; i < w + 2; i++) {
		idx1 = i + (w + 4) * 1;
		idx2 = i + (w + 4) * 2;
		idx3 = i + (w + 4) * 3;
		out->fld[idx1].ten_fld = 2 * out->fld[idx2].ten_fld - out->fld[idx3].ten_fld;

		idx1 = i + (w + 4) * 0;
		idx3 = i + (w + 4) * 4;
		out->fld[idx1].ten_fld = 2 * out->fld[idx2].ten_fld - out->fld[idx3].ten_fld;

		idx1 = i + (w + 4) * (h + 2);
		idx2 = i + (w + 4) * (h + 1);
		idx3 = i + (w + 4) * (h + 0);
		out->fld[idx1].ten_fld = 2 * out->fld[idx2].ten_fld - out->fld[idx3].ten_fld;

		idx1 = i + (w + 4) * (h + 3);
		idx3 = i + (w + 4) * (h - 1);
		out->fld[idx1].ten_fld = 2 * out->fld[idx2].ten_fld - out->fld[idx3].ten_fld;
	}



#if USE_OMP == 1
	#pragma omp sections
	{
#endif


	// top-left
#if USE_OMP == 1
		#pragma omp section
		{
#endif
		idx1 = 1 + (w + 4) * 1;
		idx2 = 2 + (w + 4) * 2;
		idx3 = 3 + (w + 4) * 3;
		out->fld[idx1].ten_fld = 2 * out->fld[idx2].ten_fld - out->fld[idx3].ten_fld;

		idx1 = 0 + (w + 4) * 0;
		idx3 = 4 + (w + 4) * 4;
		out->fld[idx1].ten_fld = 2 * out->fld[idx2].ten_fld - out->fld[idx3].ten_fld;

		idx1 = 1 + (w + 4) * 0;
		idx3 = 3 + (w + 4) * 4;
		out->fld[idx1].ten_fld = 2 * out->fld[idx2].ten_fld - out->fld[idx3].ten_fld;

		idx1 = 0 + (w + 4) * 1;
		idx3 = 4 + (w + 4) * 3;
		out->fld[idx1].ten_fld = 2 * out->fld[idx2].ten_fld - out->fld[idx3].ten_fld;

#if USE_OMP == 1
		}
#endif


	// top-right
#if USE_OMP == 1
		#pragma omp section
		{
#endif
		idx1 = -2 + (w + 4) * 2;
		idx2 = -3 + (w + 4) * 3;
		idx3 = -4 + (w + 4) * 4;
		out->fld[idx1].ten_fld = 2 * out->fld[idx2].ten_fld - out->fld[idx3].ten_fld;

		idx1 = -1 + (w + 4) * 1;
		idx3 = -5 + (w + 4) * 5;
		out->fld[idx1].ten_fld = 2 * out->fld[idx2].ten_fld - out->fld[idx3].ten_fld;

		idx1 = -2 + (w + 4) * 1;
		idx3 = -4 + (w + 4) * 5;
		out->fld[idx1].ten_fld = 2 * out->fld[idx2].ten_fld - out->fld[idx3].ten_fld;

		idx1 = -1 + (w + 4) * 2;
		idx3 = -5 + (w + 4) * 4;
		out->fld[idx1].ten_fld = 2 * out->fld[idx2].ten_fld - out->fld[idx3].ten_fld;

#if USE_OMP == 1
		}
#endif


	// bottom-left
#if USE_OMP == 1
		#pragma omp section
		{
#endif
		idx1 = 1 + (w + 4) * (h + 2);
		idx2 = 2 + (w + 4) * (h + 1);
		idx3 = 3 + (w + 4) * (h + 0);
		out->fld[idx1].ten_fld = 2 * out->fld[idx2].ten_fld - out->fld[idx3].ten_fld;

		idx1 = 0 + (w + 4) * (h + 3);
		idx3 = 4 + (w + 4) * (h - 1);
		out->fld[idx1].ten_fld = 2 * out->fld[idx2].ten_fld - out->fld[idx3].ten_fld;

		idx1 = 1 + (w + 4) * (h + 3);
		idx3 = 3 + (w + 4) * (h - 1);
		out->fld[idx1].ten_fld = 2 * out->fld[idx2].ten_fld - out->fld[idx3].ten_fld;

		idx1 = 0 + (w + 4) * (h + 2);
		idx3 = 4 + (w + 4) * (h + 0);
		out->fld[idx1].ten_fld = 2 * out->fld[idx2].ten_fld - out->fld[idx3].ten_fld;

#if USE_OMP == 1
		}
#endif


	// bottom-right
#if USE_OMP == 1
		#pragma omp section
		{
#endif
		idx1 = -2 + (w + 4) * (h + 3);
		idx2 = -3 + (w + 4) * (h + 2);
		idx3 = -4 + (w + 4) * (h + 1);
		out->fld[idx1].ten_fld = 2 * out->fld[idx2].ten_fld - out->fld[idx3].ten_fld;

		idx1 = -1 + (w + 4) * (h + 4);
		idx3 = -5 + (w + 4) * (h + 0);
		out->fld[idx1].ten_fld = 2 * out->fld[idx2].ten_fld - out->fld[idx3].ten_fld;

		idx1 = -2 + (w + 4) * (h + 4);
		idx3 = -4 + (w + 4) * (h + 0);
		out->fld[idx1].ten_fld = 2 * out->fld[idx2].ten_fld - out->fld[idx3].ten_fld;

		idx1 = -1 + (w + 4) * (h + 3);
		idx3 = -5 + (w + 4) * (h + 1);
		out->fld[idx1].ten_fld = 2 * out->fld[idx2].ten_fld - out->fld[idx3].ten_fld;

#if USE_OMP == 1
		}
#endif

#if USE_OMP == 1
	}
#endif


#if USE_OMP == 1
	}
#endif
}
