#pragma once
#include "ILayeredStrokePaintRenderer.h"
#include "IBFVRenderer_Extra.h"
#include "StrokeComparator.h"
#include "SimpleStrokeFactory.h"
#include "PaintingStyleField.h"
#include "image_pyramid.h"
#include "glew.h"
#include "GLSLManager.h"



class IBFVParam : public IPaintingParam {
public:
	unsigned char *source_img;			// origin at upper left
	unsigned char *stroke_ref_img;		// origin at lower left

	PaintingStyleField *style;
	int time;

	FieldSampleImage *edge_pixel;
	VisibleMeshInfo mesh_info;
	ImageCompositionInfo image_composition;

	// seeding method
	int seed_order;
	bool partial_ordering;


	int num_iterations;

	int embossing_type;


	IBFVParam() {
		source_img = NULL;
		stroke_ref_img = NULL;
		style = NULL;
		time = 0;
		edge_pixel = NULL;
		num_iterations = 0;
		embossing_type = 0;

		seed_order = 0;
	}
};



class IBFVRenderer : public ILayeredStrokePaintRenderer {
private:
	IBFVParam param;
	SimpleStrokeFactory s_stroke_factory;

	IBFV_SimpleLayer **layers;
	unsigned char *ref_image;
	unsigned char *stroke_ref_img_ptr;
	unsigned char *painted_image;
	float *color_noise;


	float g_stroke_spacing;
	int g_stroke_falloff;
	int g_stroke_lifetime;
	int g_stroke_fadetime;
	int g_stroke_color;
	unsigned char bgfg_threshold;


	int current_layer;
	int num_pixels;


	float *circular_stroke_coords, *circular_stroke_tex_coords, *stroke_tex_scaled_coords;
	static const float quad_vertex_coords[8];
	static const float quad_texture_coords[8];
	static const int num_circular_stroke_coords = 24;
	GLuint dl_unit_circle;
	GLuint dl_circle_tex;
	GLuint dl_draw_seed;


	static const int num_height_patterns = 3;
	static const int jitter_val_length = 10;
	GLuint stroke_shape_handle;
	GLuint stroke_tex_handle[num_height_patterns];
	float **stroke_jitter_patterns;


	unsigned char *render_buffer[4];
	unsigned char *aux[2];

	RenderTexture *initial_stroke_image;		// used for GPU optimization
	RenderTexture *render_buffer_RT[4];
	RenderTexture *tmp_target_RT;
	RenderTexture *offscreen;					// offscreen-rendering
	QuadVBO *quad;								// render off-screen texture on screen


	// used only shader flag is turned on...
	string active_shader_name;
	int shader_func_eval_params[4];

	bool use_gl_nearest_in_warp;				// for comparison purpose... 

	__int64 render_start, render_end;
	__int64 timer_freq;

public:
	IBFVRenderer(int width, int height, int num_layers);
	virtual ~IBFVRenderer();

	void setParams(IPaintingParam *param);
	bool render(unsigned char *result);
	bool resize(int width, int height);
	bool setNumLayers(int num_layers);

	size_t getNumStrokes(int layer);
	void pushSeeds(int layer, const float *opt_flow, int width, int height);
	void clearSeeds(int layer);

	double getRenderingTime();

	const unsigned char* getLayerImage(int layer);
	const unsigned char* getInitialImage(int layer);

	void setWarpingOption(bool use_gl_nearest) { use_gl_nearest_in_warp = use_gl_nearest; }
	const bool getWarpingOption() { return use_gl_nearest_in_warp; }

	int getBgFgThreshold() { return bgfg_threshold; }
	void setBgFgThreshold(unsigned char val) { bgfg_threshold = val; }

private:
	GLint prev_tex_env_mode[2];

	void initializeBuffer();
	void finalizeBuffer();

	void setUpMultiTexturePainting();
	void cleanUpMultiTexturePainting();
	void setUpMultiTexturePostProcess();
	void cleanUpMultiTexturePostProcess();

	void drawSeed(float jitter_scale, float jitter_freq, GLuint seed_tex_handle, float r, float g, float b, float a);

	void makeSeedTemplate();
	void cleanSeedTemplate();

	void scaleTexCoords(float scale);
	void restoreTexCoords();
	void copyTexCoordsList();

	void createLayeredWarpedImage();

	void (IBFVRenderer::*_renderOnLayer)();
	void renderPaintingOnLayer();
	void renderPaintingOnLayerDebug();
	void renderOneIteration();
	void selectActiveShaderAndParams();
	void passUniformVariables();

	void makeRefImage();
	void makeInitialImage();
	void regenerateStrokeList();
	void updateStrokeList();
	void seedStrokes();
	void setBackgroundColor();

	void overwriteColorBasic(unsigned char **images, unsigned char threshold, int width, int height, unsigned char *output);
	void overwriteColor3(unsigned char **images, float *weight_map, int width, int height, unsigned char *output, int op = 0);
};

