#pragma once
#include "tendes_plane.h"
#include "PaintingStyleField.h"
#include <string>
#include <fstream>
#include <vector>
#include <utility>


using namespace std;

const char *_KEYFRAME_LOADER_VERSION = "v_1.1";
const char *_KEYFRAME_VERSION_1_0 = "v_1.0";


const int _MAX_VERSION_STRING_LENGTH = 128;		// without null character



class KeyFrameLoader {
public:
	static bool save(string filename, Polyhedron &poly, PaintingStyleField &style) {
		fstream outfile;

		PaintingPreset *p;
		vector< pair<int, PaintingPreset> > style_keys;
		vector< pair<int, TransitionalPaintingPreset::INTERPOLATION_TYPE> > style_funcs;
		PaintingRegion *pr;	
		map<int, PaintingRegion*>::iterator p_itr;

		SingularElementList *se_list;
		RegularElementList *re_list;
		SingularElement *se;
		RegularElement *re;



		outfile.open(filename.c_str(), ios::out);

		// failed to create a file
		if(outfile.is_open() == false) {
			return false;
		}

		// write out a version
		outfile << _KEYFRAME_LOADER_VERSION << endl;
		outfile << endl;


		se_list = poly.singularelements;
		re_list = poly.regularelements;


		// save singular elements
		outfile << se_list->num_singularelements << endl;

		for(int i = 0; i < se_list->num_singularelements; i++) {
			se = se_list->singularelements[i];

			assert(se != NULL);				// the list should be consecutive

			outfile
				<< se->frameID				<< ' '		<< se->regionID				<< ' '
				<< (int)se->order			<< ' '		<< (int)se->type			<< ' '
				<< se->x					<< ' '		<< se->y					<< ' '
				<< se->max_dir.entry[0]		<< ' '		<< se->max_dir.entry[1]		<< ' '
				<< se->min_dir.entry[0]		<< ' '		<< se->min_dir.entry[1]		<< ' '
				<< se->max_mag				<< ' '		<< se->min_mag				<< ' '
				<< endl;
		}


		outfile << endl;

		// save regular elements
		outfile << re_list->num_regularelements << endl;

		for(int i = 0; i < re_list->num_regularelements; i++) {
			re = re_list->regularelements[i];

			assert(re != NULL);				// the list should be consecutive

			outfile
				<< re->frameID		<< ' ' 
				<< re->regionID		<< ' '  
				<< re->x			<< ' ' 
				<< re->y 			<< ' ' 
				<< re->vx 			<< ' ' 
				<< re->vy			<< ' ' 
				<< endl;
		}

		outfile << endl;
		outfile << style.getNumRegions() << endl;

		// save style parameters
		for(p_itr = style.getBegin(); p_itr != style.getEnd(); p_itr++) {
			pr = p_itr->second;

			// get enumeration of painting preset for each region
			pr->retrieveAllKeys(style_keys);

			// get all interpolations betwee keyframes
			pr->retrieveAllInterpolationFuncs(style_funcs);


			// must be the same...
			if(style_keys.size() != style_funcs.size()) {
				outfile << endl << "Error: keyframe size must be equal to interpolation func size." << endl;
				outfile.close();
				return false;
			}


			outfile << pr->getID()						<< ' ';
			outfile << (pr->getUseZeroField() ? 1 : 0)	<< ' ';
			outfile << pr->getNumKeys()					<< endl;

			// save each style parameter on a region
			for(size_t j = 0; j < style_keys.size(); j++) {
				p = &(style_keys[j].second);
			
				outfile
					<< style_keys[j].first				<< ' '  
					<< style_funcs[j].second			<< ' '  
					<< p->alpha							<< ' ' 
					<< p->area_error					<< ' ' 
					<< p->color_composition_method		<< ' ' 
					<< p->displacement					<< ' ' 
					<< p->jitter_amplitude				<< ' ' 
					<< p->num_layers					<< ' ' 
					<< p->post_process_operation		<< ' ' 
					<< p->post_process_val1				<< ' ' 
					<< p->post_process_val2				<< ' ' 
					<< p->post_process_val3				<< ' ' 
					<< p->post_process_val4				<< ' ' 
					<< p->stroke_gridsize				<< ' ' 
					<< p->stroke_initial_opacity		<< ' ' 
					<< p->stroke_max_length				<< ' ' 
					<< p->stroke_min_length				<< ' ' 
					<< p->stroke_pos_jitter				<< ' ' 
					<< p->stroke_size					<< ' ' 
					<< p->stroke_color_shift[0]			<< ' ' 
					<< p->stroke_color_shift[1]			<< ' ' 
					<< p->stroke_color_shift[2]			<< ' ' 
					<< endl;
			}
		}


		outfile << endl;

		// save gridsize for multi-style rendering
		outfile << style.getStrokeSpacingForMultiStyle() << endl;

		outfile.close();

		return true;
	}


	static bool load(string filename, Polyhedron &poly, PaintingStyleField &style) {
		fstream infile;
		char version[_MAX_VERSION_STRING_LENGTH + 1];

		infile.open(filename.c_str(), ios::in);

		// failed to create a file
		if(infile.is_open() == false) {
			return false;
		}

		infile.getline(version, _MAX_VERSION_STRING_LENGTH);
		infile.close();

		if(strcmp(version, _KEYFRAME_LOADER_VERSION) == 0) {
			//MessageBox(NULL, _T("current"), _T(""), MB_OK);
			return load_current(filename, poly, style);
		}
		else if(strcmp(version, _KEYFRAME_VERSION_1_0) == 0) {
			//MessageBox(NULL, _T("1.0"), _T(""), MB_OK);
			return load_version_1_0(filename, poly, style);
		}
		else {
			//MessageBox(NULL, _T("older"), _T(""), MB_OK);
			return load_old(filename, poly, style);
		}
	}



private:

	// load the current version
	static bool load_current(string filename, Polyhedron &poly, PaintingStyleField &style) {
		char version[_MAX_VERSION_STRING_LENGTH + 1];
		fstream infile;

		int se_order, se_type;

		int num_presets;
		PaintingPreset p;
		vector< pair<int, PaintingPreset> > style_keys;
		vector< pair<int, TransitionalPaintingPreset::INTERPOLATION_TYPE> > style_funcs;

		int num_pr;
		int region_id;
		int frame_number;
		int style_func_type;
		bool use_zero_field;
		PaintingRegion *pr;	

		int num_se;
		int num_re;
		SingularElementList *se_list;
		RegularElementList *re_list;
		SingularElement *se;
		RegularElement *re;

		float multistyle_gridsize;



		infile.open(filename.c_str(), ios::in);

		p.zero();
		style_keys.clear();
		style_funcs.clear();


		// failed to create a file
		if(infile.is_open() == false) {
			return false;
		}


		// read and discard
		infile.getline(version, _MAX_VERSION_STRING_LENGTH);


		poly.singularelements->finalize();
		poly.regularelements->finalize();
		poly.singularities->finalize();
		poly.singularelements = new SingularElementList(20);
		poly.regularelements = new RegularElementList(20);
		poly.singularities = new SingularityList(20);


		se_list = poly.singularelements;
		re_list = poly.regularelements;


		// read singular elements
		infile >> num_se;

		for(int i = 0; i < num_se; i++) {
			se = new SingularElement(poly.vec_ten_mode, 0, 0, 0, 0);

			infile
				>> se->frameID			>> se->regionID 
				>> se_order				>> se_type 
				>> se->x				>> se->y 
				>> se->max_dir.entry[0]	>> se->max_dir.entry[1]	
				>> se->min_dir.entry[0]	>> se->min_dir.entry[1]	
				>> se->max_mag			>> se->min_mag 
				;


			se->order = (unsigned char)se_order;
			se->type = (unsigned char)se_type;

			if(se->regionID == -1 || style.getRegion(se->regionID) != NULL) {
				se->Construct_mat();
				se->calc_display_attributes();
				se_list->append(se);
			}
			else {					// no region exists for this element
				delete se;
			}
		}


		// read regular elements
		infile >> num_re;

		for(int i = 0; i < num_re; i++) {
			re = new RegularElement(0, 0);

			infile
				>> re->frameID 
				>> re->regionID 
				>> re->x 
				>> re->y 
				>> re->vx 
				>> re->vy 
				;

			if(re->regionID == -1 || style.getRegion(re->regionID) != NULL) {
				re->calc_display_attributes();
				re_list->append(re);
			}
			else {
				delete re;
			}
		}


		// read the number of regions
		infile >> num_pr;

		// read style parameters
		for(int i = 0; i < num_pr; i++) {
			infile >> region_id;
			infile >> use_zero_field;
			infile >> num_presets;
			style_keys.clear();
			style_funcs.clear();

			// filter only valid elements
			for(int j = 0; j < num_presets; j++) {
				infile
					>> frame_number 
					>> style_func_type 
					>> p.alpha 
					>> p.area_error 
					>> p.color_composition_method 
					>> p.displacement 
					>> p.jitter_amplitude 
					>> p.num_layers 
					>> p.post_process_operation 
					>> p.post_process_val1	
					>> p.post_process_val2	
					>> p.post_process_val3	
					>> p.post_process_val4	
					>> p.stroke_gridsize 
					>> p.stroke_initial_opacity 
					>> p.stroke_max_length	
					>> p.stroke_min_length	
					>> p.stroke_pos_jitter	
					>> p.stroke_size
					>> p.stroke_color_shift[0]
					>> p.stroke_color_shift[1]	
					>> p.stroke_color_shift[2]
					;


				if(0 <= frame_number && frame_number < style.getNumFrames()) {
					style_keys.push_back(pair<int, PaintingPreset>(frame_number, p));
					style_funcs.push_back(pair<int, TransitionalPaintingPreset::INTERPOLATION_TYPE>(frame_number, (TransitionalPaintingPreset::INTERPOLATION_TYPE)style_func_type));
				}
			}

			if((pr = style.getRegion(region_id)) != NULL) {
				pr->setAllKeys(style_keys);
				pr->setAllInterpolationFuncs(style_funcs);
				pr->setUseZeroField(use_zero_field);
			}
		}


		infile >> multistyle_gridsize;
		style.setStrokeSpacingForMultiStyle(multistyle_gridsize);


		infile.close();

		return true;
	}


	static bool load_version_1_0(string filename, Polyhedron &poly, PaintingStyleField &style) {
		char version[_MAX_VERSION_STRING_LENGTH + 1];
		fstream infile;

		int se_order, se_type;

		int num_presets;
		PaintingPreset p;
		vector< pair<int, PaintingPreset> > style_keys;

		int num_pr;
		int region_id;
		int frame_number;
		bool use_zero_field;
		PaintingRegion *pr;	

		int num_se;
		int num_re;
		SingularElementList *se_list;
		RegularElementList *re_list;
		SingularElement *se;
		RegularElement *re;

		float multistyle_gridsize;



		infile.open(filename.c_str(), ios::in);

		p.zero();
		style_keys.clear();


		// failed to create a file
		if(infile.is_open() == false) {
			return false;
		}


		// read and discard
		infile.getline(version, _MAX_VERSION_STRING_LENGTH);


		poly.singularelements->finalize();
		poly.regularelements->finalize();
		poly.singularities->finalize();
		poly.singularelements = new SingularElementList(20);
		poly.regularelements = new RegularElementList(20);
		poly.singularities = new SingularityList(20);


		se_list = poly.singularelements;
		re_list = poly.regularelements;


		// read singular elements
		infile >> num_se;

		for(int i = 0; i < num_se; i++) {
			se = new SingularElement(poly.vec_ten_mode, 0, 0, 0, 0);

			infile
				>> se->frameID			>> se->regionID 
				>> se_order				>> se_type 
				>> se->x				>> se->y 
				>> se->max_dir.entry[0]	>> se->max_dir.entry[1]	
				>> se->min_dir.entry[0]	>> se->min_dir.entry[1]	
				>> se->max_mag			>> se->min_mag 
				;


			se->order = (unsigned char)se_order;
			se->type = (unsigned char)se_type;

			if(se->regionID == -1 || style.getRegion(se->regionID) != NULL) {
				se->Construct_mat();
				se->calc_display_attributes();
				se_list->append(se);
			}
			else {					// no region exists for this element
				delete se;
			}
		}


		// read regular elements
		infile >> num_re;

		for(int i = 0; i < num_re; i++) {
			re = new RegularElement(0, 0);

			infile
				>> re->frameID 
				>> re->regionID 
				>> re->x 
				>> re->y 
				>> re->vx 
				>> re->vy 
				;

			if(re->regionID == -1 || style.getRegion(re->regionID) != NULL) {
				re->calc_display_attributes();
				re_list->append(re);
			}
			else {
				delete re;
			}
		}


		// read the number of regions
		infile >> num_pr;

		// read style parameters
		for(int i = 0; i < num_pr; i++) {
			infile >> region_id;
			infile >> use_zero_field;
			infile >> num_presets;
			style_keys.clear();

			// filter only valid elements
			for(int j = 0; j < num_presets; j++) {
				infile
					>> frame_number 
					>> p.alpha 
					>> p.area_error 
					>> p.color_composition_method 
					>> p.displacement 
					>> p.jitter_amplitude 
					>> p.num_layers 
					>> p.post_process_operation 
					>> p.post_process_val1	
					>> p.post_process_val2	
					>> p.post_process_val3	
					>> p.post_process_val4	
					>> p.stroke_gridsize 
					>> p.stroke_initial_opacity 
					>> p.stroke_max_length	
					>> p.stroke_min_length	
					>> p.stroke_pos_jitter	
					>> p.stroke_size
					>> p.stroke_color_shift[0]
					>> p.stroke_color_shift[1]	
					>> p.stroke_color_shift[2]
					;


				if(0 <= frame_number && frame_number < style.getNumFrames()) {
					style_keys.push_back(pair<int, PaintingPreset>(frame_number, p));
				}
			}

			if((pr = style.getRegion(region_id)) != NULL) {
				pr->setAllKeys(style_keys);
				pr->setUseZeroField(use_zero_field);
			}
		}


		infile >> multistyle_gridsize;
		style.setStrokeSpacingForMultiStyle(multistyle_gridsize);


		infile.close();

		return true;
	}


	// load old version
	static bool load_old(string filename, Polyhedron &poly, PaintingStyleField &style) {
		fstream infile;

		int se_order, se_type;

		int num_presets;
		PaintingPreset p;
		vector< pair<int, PaintingPreset> > style_keys;

		int num_pr;
		int region_id;
		int frame_number;
		bool use_zero_field;
		PaintingRegion *pr;	

		int num_se;
		int num_re;
		SingularElementList *se_list;
		RegularElementList *re_list;
		SingularElement *se;
		RegularElement *re;

		float multistyle_gridsize;



		infile.open(filename.c_str(), ios::in);

		p.zero();
		style_keys.clear();


		// failed to create a file
		if(infile.is_open() == false) {
			return false;
		}


		poly.singularelements->finalize();
		poly.regularelements->finalize();
		poly.singularities->finalize();
		poly.singularelements = new SingularElementList(20);
		poly.regularelements = new RegularElementList(20);
		poly.singularities = new SingularityList(20);


		se_list = poly.singularelements;
		re_list = poly.regularelements;


		// read singular elements
		infile >> num_se;

		for(int i = 0; i < num_se; i++) {
			se = new SingularElement(poly.vec_ten_mode, 0, 0, 0, 0);

			infile
				>> se->frameID			>> se->regionID 
				>> se_order				>> se_type 
				>> se->x				>> se->y 
				>> se->max_dir.entry[0]	>> se->max_dir.entry[1]	
				>> se->min_dir.entry[0]	>> se->min_dir.entry[1]	
				>> se->max_mag			>> se->min_mag 
				;


			se->order = (unsigned char)se_order;
			se->type = (unsigned char)se_type;

			if(se->regionID == -1 || style.getRegion(se->regionID) != NULL) {
				se->Construct_mat();
				se->calc_display_attributes();
				se_list->append(se);
			}
			else {					// no region exists for this element
				delete se;
			}
		}


		// read regular elements
		infile >> num_re;

		for(int i = 0; i < num_re; i++) {
			re = new RegularElement(0, 0);

			infile
				>> re->frameID 
				>> re->regionID 
				>> re->x 
				>> re->y 
				>> re->vx 
				>> re->vy 
				;

			if(re->regionID == -1 || style.getRegion(re->regionID) != NULL) {
				re->calc_display_attributes();
				re_list->append(re);
			}
			else {
				delete re;
			}
		}


		// read the number of regions
		infile >> num_pr;

		if(infile.eof() == true) {
			return true;
		}


		// read style parameters
		for(int i = 0; i < num_pr; i++) {
			infile >> region_id;
			infile >> use_zero_field;
			infile >> num_presets;
			style_keys.clear();

			// filter only valid elements
			for(int j = 0; j < num_presets; j++) {
				infile
					>> frame_number 
					>> p.alpha 
					>> p.area_error 
					>> p.color_composition_method 
					>> p.displacement 
					>> p.jitter_amplitude 
					>> p.num_layers 
					>> p.post_process_operation 
					>> p.post_process_val1	
					>> p.post_process_val2	
					>> p.post_process_val3	
					>> p.post_process_val4	
					>> p.stroke_gridsize 
					>> p.stroke_initial_opacity 
					>> p.stroke_max_length	
					>> p.stroke_min_length	
					>> p.stroke_pos_jitter	
					>> p.stroke_size
					;

				if(0 <= frame_number && frame_number < style.getNumFrames()) {
					style_keys.push_back(pair<int, PaintingPreset>(frame_number, p));
				}
			}

			if((pr = style.getRegion(region_id)) != NULL) {
				pr->setAllKeys(style_keys);
				pr->setUseZeroField(use_zero_field);
			}
		}


		infile >> multistyle_gridsize;

		// this branch was introduced for compativility with old control files.
		// feel free to modify unless you want to retain the old format.
		if(infile.eof() == false) {
			style.setStrokeSpacingForMultiStyle(multistyle_gridsize);
		}
		else {
			style.setStrokeSpacingForMultiStyle(PaintingStyleField::mp_default_stroke_spacing);
		}

		infile.close();

		return true;
	}
};
